package io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.Offset
import io.tswf.groovy.bettergroovy.asm.transmutation.annotation.Transmute
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.*
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.InjectionOffset
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.points.TransmutationInjectionPointManager
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.reader.remap.TransmutationRemapper
import io.tswf.groovy.bettergroovy.asm.transmutation.meta.MissingTargetStrategy
import io.tswf.groovy.bettergroovy.asm.transmutation.meta.TransmutationMethodInjectionPointMetadata
import io.tswf.groovy.bettergroovy.asm.transmutation.meta.TransmutationMethodMetadata
import io.tswf.groovy.bettergroovy.asm.transmutation.util.ClassReferenceUtil
import io.tswf.groovy.bettergroovy.asm.transmutation.util.CompositeAtUtil
import io.tswf.groovy.bettergroovy.asm.transmutation.util.MethodDescriptionUtil
import org.objectweb.asm.Type
import org.objectweb.asm.tree.AnnotationNode
import org.objectweb.asm.tree.ClassNode
import org.objectweb.asm.tree.MethodNode

import java.util.stream.Stream

@CompileStatic
class TransmutationMetadataReader {

    public static final String TARGET_CLASS_PARAMETER_NAME = "targetClass"
    public static final String TARGET_METHOD_PARAMETER_NAME = "targetMethod"

    @Lazy
    static List<IReturnValueProcessorReader> returnValueProcessorReaders = ServiceLoader.load(IReturnValueProcessorReader).toList()

    @Lazy
    static List<IPreInjectionHooksReader> preInjectionHooksReaders = ServiceLoader.load(IPreInjectionHooksReader).toList()
    
    @Lazy
    static List<IPostInjectionHooksReader> postInjectionHooksReaders = ServiceLoader.load(IPostInjectionHooksReader).toList()

    @Lazy
    static List<ICustomParameterLoaderReader> customParameterLoaderReaders = ServiceLoader.load(ICustomParameterLoaderReader).toList()

    static TransmutationMethodMetadata readTransmutationMetadata(ClassNode owner, MethodNode transmutationMethod) {
        try {
            def resultTransmutationMetadata = new TransmutationMethodMetadata()

            resultTransmutationMetadata.transmutationOwner = owner
            resultTransmutationMetadata.transmutationMethod = transmutationMethod

            def transmutationAnnotation = transmutationMethod.getAnnotationByType(Transmute)

            // Читаем простые парамеры

            def onMissingStrategy = transmutationAnnotation.getValueByName("onMissing", MissingTargetStrategy) { MissingTargetStrategy.WARNING }

            resultTransmutationMetadata.targetMissingStrategy = onMissingStrategy

            // Читаем точки инъекции из параметров 'at' и 'dependedAt' аннотации трансмутации
            CompositeAtUtil.findInjectionPointsComposite(transmutationAnnotation)?.with {
                resultTransmutationMetadata.injectionPoints.addAll(it)
            }

            // Читаем точки инъекции из аннотаций самого метода
            Stream.of(transmutationMethod.visibleAnnotations, transmutationMethod.invisibleAnnotations)
                .filter { it != null }
                .flatMap { it.stream() }
                .map {
                    TransmutationInjectionPointManager.findInjectionPointReaderByMarkerDesc(it.desc)
                        ?.with { injectionPointReader ->
                            def injectionPointReadingContext = new InjectionPointReadingContext(annotationNodeFromRead: it)
                            injectionPointReader.handleInjectionPointReading(injectionPointReadingContext)
                            return injectionPointReadingContext.result
                        }
                }
                .filter { it != null }
                .toList()
                .let { foundInjectionPointsOnMethod ->
                    if (foundInjectionPointsOnMethod.size() > 1) {
                        throw new MultipleInjectionPointsFoundException("There is multiple injection points found on ${transmutationMethod.javaFullName} in ${owner.javaName}! It is not supported now! (${foundInjectionPointsOnMethod}")
                    }

                    def offset = transmutationMethod.findAnnotationByType(Offset)
                            ?.with {
                                it.getValueByName("value", InjectionOffset)
                            }
                            ?: InjectionOffset.DEFAULT


                    foundInjectionPointsOnMethod.each { injectionPoint ->
                        def injectionPointMetadata = new TransmutationMethodInjectionPointMetadata(
                                injectionOffset: offset,
                                injectionPoint: injectionPoint
                        )

                        resultTransmutationMetadata.injectionPoints << injectionPointMetadata
                    }
                }

            // Читаем целевой класс
            def rawTargetClass = transmutationAnnotation.getValueByName(TARGET_CLASS_PARAMETER_NAME, AnnotationNode).with { targetClassRef ->
                ClassReferenceUtil.getTypeFromClassRefAnnotation(targetClassRef)
            }

            // Для чтения остальной части трансмутации оставляем сырую версию, ниже заменим на ремап.
            resultTransmutationMetadata.targetClass = rawTargetClass

            // Читаем описание целевого метода
            def targetMethodDescriptionAnnotation = transmutationAnnotation.getValueByName(TARGET_METHOD_PARAMETER_NAME, AnnotationNode)
            def rawTargetMethodDescription = MethodDescriptionUtil.findMethodFromAnnotation(targetMethodDescriptionAnnotation)
            def mappedTargetMethodDescription = TransmutationRemapper.mapMethod(rawTargetClass, rawTargetMethodDescription)

            resultTransmutationMetadata.targetMethodDescription = mappedTargetMethodDescription

            def transmutationMethodParameterTypes = transmutationMethod.parameterTypes

            // Если есть кастомные параметры (а они всегда находятся ПОСЛЕ обычных - читаем для них кастомные загрузчики)
            def customParametersCount = transmutationMethodParameterTypes.size() - mappedTargetMethodDescription.argumentTypes.length
            if (customParametersCount > 0) {
                for (def customParameterIndex = mappedTargetMethodDescription.argumentTypes.length; customParameterIndex < transmutationMethodParameterTypes.size(); customParameterIndex ++) {
                    def customParameterReadingContext = new CustomParameterLoaderReadingContext(
                            transmutationMetadata: resultTransmutationMetadata,
                            parameterIndex: customParameterIndex,
                            parameterType: transmutationMethod.getParameterType(customParameterIndex)
                    )

                    for (def customParameterReader : customParameterLoaderReaders) {
                        try {
                            customParameterReader.handleLoaderSearch(customParameterReadingContext)
                            if (customParameterReadingContext.completed) {
                                break
                            }
                        } catch (Throwable e) {
                            throw new CustomParameterLoaderReadingException("Exception occurred while trying to load custom parameter with index ${customParameterIndex} by reader ${customParameterReader}", e)
                        }
                    }

                    if (!customParameterReadingContext.completed) {
                        throw new InvalidTransmutationDescriptionException("There is no custom parameter loader for ${customParameterIndex + 1}'st parameter in transmutation method ${transmutationMethod.javaFullName} of ${owner.javaName}!")
                    }
                }
            }

            // Загружаем хуки пре и пост инъекции

            def preInjectionHooksReadingContext = new PreInjectionHooksReadingContext(
                    transmutationMetadata: resultTransmutationMetadata
            )

            for (def preInjectionHooksReader : preInjectionHooksReaders) {
                try {
                    preInjectionHooksReader.handlePreInjectionHooksReading(preInjectionHooksReadingContext)
                } catch (Throwable e) {
                    throw new PreInjectionHooksReadingException("Exception occurred while reading pre injection hooks of ${transmutationMethod.javaFullName} of ${owner.javaName} by ${preInjectionHooksReader}", e)
                }
            }

            def returnValueProcessorReadingContext = new ReturnValueProcessorReadingContext(
                    transmutationMetadata: resultTransmutationMetadata
            )

            for (def returnValueProcessorReader : returnValueProcessorReaders) {
                try {
                    returnValueProcessorReader.handleReturnValueProcessorReading(returnValueProcessorReadingContext)
                } catch (Throwable e) {
                    throw new ReturnValueProcessorReadingException("Exception occurred while trying to read return value processor for ${transmutationMethod.javaFullName} of ${owner.javaName} by ${returnValueProcessorReader}", e)
                }
            }

            if (returnValueProcessorReadingContext.readedProcessors.isEmpty()) {
                throw new NoReturnValueProcessorDescribedException("There is no one annotation selects return value handle strategy in ${transmutationMethod.javaFullName} of ${owner.javaName}")
            }

            if (returnValueProcessorReadingContext.readedProcessors.size() > 1) {
                throw new MultipleReturnValueProcessorsFoundException("There is more than one return value processor found on ${transmutationMethod.javaFullName} of ${owner.javaName}")
            }

            resultTransmutationMetadata.returnValueProcessor = returnValueProcessorReadingContext.readedProcessors.first()

            def postInjectionHooksReadingContext = new PostInjectionHooksReadingContext(
                    transmutationMetadata: resultTransmutationMetadata
            )

            for (def postInjectionHooksReader : postInjectionHooksReaders) {
                try {
                    postInjectionHooksReader.handlePostInjectionHooksReading(postInjectionHooksReadingContext)
                } catch (Throwable e) {
                    throw new PostInjectionHooksReadingException("Exception occurred while reading post injection hooks of ${transmutationMethod.javaFullName} of ${owner.javaName} by ${postInjectionHooksReader}", e)
                }
            }

            // Ремапим только под конец для того, чтобы остальные читалки могли использовать сырой тип для своего ремаппинга
            def mappedTargetClass = TransmutationRemapper.mapType(rawTargetClass)
            resultTransmutationMetadata.targetClass = mappedTargetClass

            return resultTransmutationMetadata
        } catch (Throwable e) {
            throw new TransmutationReadingException("Exception occurred while trying to read transmutation method ${transmutationMethod.javaFullName} of ${owner.javaName}", e)
        }
    }
}
