package io.tswf.groovy.bettergroovy.asm.transmutation.injection.returnvalue

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationException
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.UnknownLabelToJumpNameException
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.ITransmutationMethodReturnValueProcessor
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.TransmutationInjectionToMethodAtInsnContext
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.hook.DefaultHookConstants
import org.objectweb.asm.Opcodes
import org.objectweb.asm.Type
import org.objectweb.asm.tree.InsnList
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.JumpInsnNode
import org.objectweb.asm.tree.LabelNode
import org.objectweb.asm.tree.LdcInsnNode
import org.objectweb.asm.tree.MethodInsnNode
import org.objectweb.asm.tree.TypeInsnNode
import org.objectweb.asm.tree.VarInsnNode

import static io.tswf.groovy.bettergroovy.asm.transmutation.injection.hook.DefaultHookConstants.NEW_CREATED_LABELS

import static org.objectweb.asm.Opcodes.*

@CompileStatic
class JumpToCreatedLabelReturnValueProcessor implements ITransmutationMethodReturnValueProcessor {

    static Type defaultUnknownLabelNameExceptionType = UnknownLabelToJumpNameException.asAsmType()

    Type unknownLabelNameExceptionType = defaultUnknownLabelNameExceptionType

    @Override
    void handleTransmutationMethodReturnValue(TransmutationInjectionToMethodAtInsnContext context) {
        def createdLabelsMap = (Map<String, LabelNode>) context.allPointsContext.customPayload.getByKey(NEW_CREATED_LABELS)

        if (createdLabelsMap.isNullOrEmpty()) {
            throw new InvalidTransmutationException("There is no created labels found for jumping after method returns!")
        }

        def targetMethod = context.allPointsContext.targetMethod
        def transmutationMethod = context.allPointsContext.transmutationMetadata.transmutationMethod
        def transmutationMethodReturnType = transmutationMethod.returnType

        // Проверяем, что метод трансмутации отдает String
        if (transmutationMethodReturnType.descriptor != String.asmDescriptor) {
            throw new InvalidTransmutationException("The transmutation method ${transmutationMethod.javaFullName} of ${context.allPointsContext.transmutationMetadata.transmutationOwner.javaName} return type is ${transmutationMethodReturnType.descriptor}, but must be a ${String.asmDescriptor} for label jumping!")
        }

        // Создаем лейбл, который разместим после всей обработки на случай, если трансмутация вернет null имя лейбла
        def labelIfNull = new LabelNode()

        def jumpLabelLocalVarId = context.allPointsContext.customPayload.computeIfAbsent("jump-label-transmutation-return-local-var-holder-id") {
            targetMethod.createNewLocal(String, "labelToJump", true)
        }

        // Проверяем на null и прыгаем к концу обработки, если null. В этом случае никаких особых действий в методе-цели не произойдет
        context.insnsToInsertToCurrent << new InsnNode(DUP)
        context.insnsToInsertToCurrent << new VarInsnNode(ASTORE, jumpLabelLocalVarId)
        context.insnsToInsertToCurrent << new JumpInsnNode(IFNULL, labelIfNull)

        // Для каждого созданного лейбла делаем проверку на то, что трансмутация вернула его имя и прыгаем к нему, если это так.
        for (def labelEntry : createdLabelsMap.entrySet()) {
            context.insnsToInsertToCurrent << new VarInsnNode(ALOAD, jumpLabelLocalVarId)
            context.insnsToInsertToCurrent << new LdcInsnNode(labelEntry.key.toString())
            context.insnsToInsertToCurrent << new MethodInsnNode(INVOKESTATIC, Objects.asmInternalName, "equals", "(Ljava/lang/Object;Ljava/lang/Object;)Z", false)

            def labelToJump = labelEntry.value

            // Если мы прыгаем перед RETURN или ATHROW, то значение со стека надо убрать. В следующий приход оно снова положится.
            // Для этого над целевым лейблом мы создаем еще один, к которому и направим прыжок.
            def popBeforeJumpOpcode = getPopValueBeforeJumpOpcode(context)
            if (popBeforeJumpOpcode > -1) {
                def labelWithPop = new LabelNode()
                def popInsns = new InsnList()
                popInsns << new JumpInsnNode(GOTO, labelEntry.value)
                popInsns << labelWithPop
                popInsns << new InsnNode(popBeforeJumpOpcode)
                targetMethod.instructions.insertBefore(labelEntry.value, popInsns)
                labelToJump = labelWithPop
            }

            context.insnsToInsertToCurrent << new JumpInsnNode(IFNE, labelToJump)
        }

        // Если трансмутация вернула не null и не имя одного из созданных лейблов, то это ошибка, которую надо выбросить в рантайме.
        context.insnsToInsertToCurrent << new TypeInsnNode(NEW, unknownLabelNameExceptionType.internalName)
        context.insnsToInsertToCurrent << new InsnNode(DUP)
        context.insnsToInsertToCurrent << new LdcInsnNode(createdLabelsMap.keySet().toString())
        context.insnsToInsertToCurrent << new VarInsnNode(ALOAD, jumpLabelLocalVarId)
        context.insnsToInsertToCurrent << new MethodInsnNode(INVOKESTATIC, JumpToCreatedLabelReturnValueProcessor.asmInternalName, "getUnknownLabelExceptionText", "(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;")

        context.insnsToInsertToCurrent << new MethodInsnNode(INVOKESPECIAL, unknownLabelNameExceptionType.internalName, "<init>", "(Ljava/lang/String;)V", false)
        context.insnsToInsertToCurrent << new InsnNode(ATHROW)

        // Если трансмутация вернула null, то никаких прыжков не будет совершено.
        context.insnsToInsertToCurrent << labelIfNull
    }

    int getPopValueBeforeJumpOpcode(TransmutationInjectionToMethodAtInsnContext context) {
        def targetMethod = context.allPointsContext.targetMethod
        def targetMethodReturnType = targetMethod.returnType
        def isReturnsAndNotVoid = Opcodes.isOpcodeReturn(context.startInsn.opcode) && targetMethodReturnType != Type.VOID_TYPE
        def isThrows = context.startInsn.opcode == ATHROW

        def isReturnValueStoredInLocalVar = context.allPointsContext.customPayload.getByKey(DefaultHookConstants.RETURN_LOCAL_VAR_INDEX_KEY) != null
        def isThrownExceptionStoredInLocalVar = context.allPointsContext.customPayload.getByKey(DefaultHookConstants.THROWN_EXCEPTION_LOCAL_VAR_KEY) != null

        if (isReturnsAndNotVoid && !isReturnValueStoredInLocalVar) {
            return targetMethodReturnType.size == 1 ? POP : POP2
        }

        if (isThrows && !isThrownExceptionStoredInLocalVar) {
            return POP
        }

        return -1
    }

    /** Вызывается чтобы сконкатить имя неизвестного лейбла и спиок известных из класса-цели */
    static String getUnknownLabelExceptionText(String labelName, String knownLabels) {
        return "Unknown label name to jump: '${labelName}'! Available labels here: ${knownLabels}"
    }
}
