package io.tswf.groovy.bettergroovy.asm.transmutation.injection.returnvalue

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InjectionPointCantBeReachedException
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.InvalidTransmutationException
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.LockedLocalVariableException
import io.tswf.groovy.bettergroovy.asm.transmutation.exception.TransmutationReturnValueCastException
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.ITransmutationMethodReturnValueProcessor
import io.tswf.groovy.bettergroovy.asm.transmutation.injection.contexts.TransmutationInjectionToMethodAtInsnContext
import org.objectweb.asm.Type
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.JumpInsnNode
import org.objectweb.asm.tree.LabelNode
import org.objectweb.asm.tree.LdcInsnNode
import org.objectweb.asm.tree.LocalVariableNode
import org.objectweb.asm.tree.MethodInsnNode
import org.objectweb.asm.tree.TryCatchBlockNode
import org.objectweb.asm.tree.TypeInsnNode
import org.objectweb.asm.tree.VarInsnNode

import java.lang.reflect.Modifier

import static org.objectweb.asm.Opcodes.*

@CompileStatic
class ReturningIntoLocalVariableTransmutationReturnValueProcessor implements ITransmutationMethodReturnValueProcessor {

    static Type defaultCastExceptionWrapperClassType = TransmutationReturnValueCastException.asAsmType()
    Type castExceptionWrapperClassType = defaultCastExceptionWrapperClassType
    Type localVariableType
    int localVariableIndex
    boolean force

    ReturningIntoLocalVariableTransmutationReturnValueProcessor(Type localVariableType, int localVariableIndex, boolean force) {
        this.localVariableType = localVariableType
        this.localVariableIndex = localVariableIndex
        this.force = force
    }

    @Override
    void handleTransmutationMethodReturnValue(TransmutationInjectionToMethodAtInsnContext context) {
        def transmutationMethod = context.allPointsContext.transmutationMetadata.transmutationMethod
        def targetMethod = context.allPointsContext.targetMethod
        def targetMethodIsStatic = Modifier.isStatic(targetMethod.access)

        // Если происходит попытка записать 0 локальную переменную на нестатическом методе, то это this и так нельзя.
        if (!targetMethodIsStatic && localVariableIndex == 0) {
            throw new InvalidTransmutationException("Can't rewrite 'this' local variable in target method ${targetMethod.javaFullName} of ${context.allPointsContext.targetMethodOwner.javaName} by ${transmutationMethod.javaFullName} of ${context.allPointsContext.transmutationMetadata.transmutationOwner.javaName}!")
        }

        def targetLocalVariable = targetMethod.localVariables
            .find { it.index == localVariableIndex && it.desc == localVariableType.descriptor }
            .ensureNotNull {
                new InjectionPointCantBeReachedException(
                        "There is no local variable with index ${localVariableIndex} and type ${localVariableType} in ${targetMethod.javaFullName} of ${context.allPointsContext.targetMethodOwner.javaName}! Available variables: ${getAvailableLocalVariablesFancyString(targetMethod.localVariables)}"
                )
            }

        def indexToSearch = targetLocalVariable.index
        if (!targetMethodIsStatic) {
            indexToSearch --
        }

        def parameterWithSameIndex = targetMethod.parameters?.get(indexToSearch)
        if (parameterWithSameIndex != null) {
            if (Modifier.isFinal(parameterWithSameIndex.access)) {
                if (!force) {
                    throw new LockedLocalVariableException("Target local variable (index: ${localVariableIndex}, type: ${localVariableType.className}) is final. Use force flag to overwrite it if you sure!")
                } else {
                    parameterWithSameIndex.access &= ~Modifier.FINAL
                }
            }
        }

        def transmutationMethodReturnType = transmutationMethod.returnType

        // Сравниваем типы возвращаемого трансмутацией значения и типа локальной переменной. Если они не совпадают, то пробуем сделать каст или кидаем ошибку.
        if (transmutationMethodReturnType != localVariableType) {
            if (transmutationMethodReturnType == Object.asAsmType() && !localVariableType.isPrimitive()) {
                // Если метод трансмутации отдает Object, а тип нужной локальной переменной не примитивный, то можно попробовать сделать каст возвращенного значения к типу целевого

                // Для этого создаем try-catch блок, в котором будем работать и в случае ошибки бросать исключение, которое объяснит по вине какой именно трансмутации ошибка.
                def caughtExceptionLocalVarIndex = targetMethod.createNewLocal(Throwable, "cast_exception", true)
                def localVariableNewValueLocalVarIndex = targetMethod.createNewLocal(localVariableType, "local_new_value", true)

                def labelStartCast = new LabelNode()
                def labelEndCast = new LabelNode()

                def labelStartCatch = new LabelNode()
                def labelAfterCast = new LabelNode()

                def castTryCatchBlock = new TryCatchBlockNode(labelStartCast, labelEndCast, labelStartCatch, Throwable.asmInternalName)

                // Работаем с объектными инструкциями потому, что примитивные типы не способны вызвать генерацию этого блока кода
                context.insnsToInsertToCurrent << new VarInsnNode(ASTORE, localVariableNewValueLocalVarIndex)

                // Загружаем значение, возвращенное трансмутацией на стек и пробуем сделать ему каст
                context.insnsToInsertToCurrent << labelStartCast
                context.insnsToInsertToCurrent << new VarInsnNode(ALOAD, localVariableNewValueLocalVarIndex)
                context.insnsToInsertToCurrent << new TypeInsnNode(CHECKCAST, localVariableType.internalName)

                // Если каст успешен, то просто переходим дальше
                context.insnsToInsertToCurrent << new JumpInsnNode(GOTO, labelAfterCast)

                context.insnsToInsertToCurrent << labelEndCast

                // Если каст завершился ошибкой, то забираем со стека выброшенное исключение и сохраняем его в локальную переменную
                context.insnsToInsertToCurrent << labelStartCatch
                context.insnsToInsertToCurrent << new VarInsnNode(ASTORE, caughtExceptionLocalVarIndex)

                // Создаем инстанс исключения-обертки, который будет описывать что за трансмутация вернула некорректное значение и какое.
                context.insnsToInsertToCurrent << new TypeInsnNode(NEW, castExceptionWrapperClassType.internalName)
                context.insnsToInsertToCurrent << new InsnNode(DUP)
                context.insnsToInsertToCurrent << new VarInsnNode(ALOAD, caughtExceptionLocalVarIndex)
                context.insnsToInsertToCurrent << new VarInsnNode(ALOAD, localVariableNewValueLocalVarIndex)
                context.insnsToInsertToCurrent << new LdcInsnNode("${transmutationMethod.javaFullName} of ${context.allPointsContext.targetMethodOwner.asAsmType().className}".toString())
                context.insnsToInsertToCurrent << new MethodInsnNode(INVOKESPECIAL, castExceptionWrapperClassType.internalName, "<init>", "(Ljava/lang/Throwable;Ljava/lang/Object;Ljava/lang/String;)V", false)
                context.insnsToInsertToCurrent << new InsnNode(ATHROW)

                // Размещаем лейбл после блока catch для того, чтобы успешные касты обходили его стороной
                context.insnsToInsertToCurrent << labelAfterCast

                // Добавляем информацию о блоке try-catch в метод
                if (targetMethod.tryCatchBlocks == null) {
                    targetMethod.tryCatchBlocks = new ArrayList<>()
                }

                targetMethod.tryCatchBlocks << castTryCatchBlock

            } else {
                // А если у нас ситуация, когда каст делать нет смысла - бросаем ошибку
                throw new InvalidTransmutationException("Target local variable type ${localVariableType.className} of method ${targetMethod.javaFullName} of ${context.allPointsContext.targetMethodOwner.javaName} does not matches with transmutation method return type ${transmutationMethodReturnType.className} from ${transmutationMethod.javaFullName} of ${context.allPointsContext.transmutationMetadata.transmutationOwner.javaName}!")
            }
        }

        context.insnsToInsertToCurrent << new VarInsnNode(localVariableType.getOpcode(ISTORE), localVariableIndex)
    }

    String getAvailableLocalVariablesFancyString(List<LocalVariableNode> vars) {
        return vars
            .map { var -> "${var.index}: ${Type.getType(var.desc).className}" }
            .toString()
    }
}
